# -*- coding: utf-8 -*-
#
import logging
import sys
from os import getenv, path
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

import rdflib
from rdflib import BNode, Literal, URIRef

from .consts import (
    RDF_type,
    SH_conforms,
    SH_result,
    SH_ValidationReport,
    env_truths,
)
from .errors import ReportableRuntimeError
from .extras import check_extra_installed
from .functions import apply_functions, gather_functions, unapply_functions
from .pytypes import GraphLike, SHACLExecutor
from .rdfutil import (
    add_baked_in,
    clone_blank_node,
    clone_graph,
    inoculate,
    inoculate_dataset,
    mix_datasets,
    mix_graphs,
)
from .rules import apply_rules, gather_rules
from .run_type import PySHACLRunType
from .shapes_graph import ShapesGraph
from .target import apply_target_types, gather_target_types

USE_FULL_MIXIN = getenv("PYSHACL_USE_FULL_MIXIN") in env_truths


class Validator(PySHACLRunType):
    def __init__(
        self,
        data_graph: GraphLike,
        *args,
        shacl_graph: Optional[GraphLike] = None,
        ont_graph: Optional[GraphLike] = None,
        options: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        options = options or {}
        self._load_default_options(options)
        self.options = options  # type: dict
        self.logger = options['logger']  # type: logging.Logger
        self.debug = options['debug']
        self.pre_inferenced = kwargs.pop('pre_inferenced', False)
        self.inplace = options['inplace']
        if not isinstance(data_graph, rdflib.Graph):
            raise RuntimeError("data_graph must be a rdflib Graph object")
        self.data_graph = data_graph  # type: GraphLike
        self._target_graph: Union[GraphLike, None] = None
        self.ont_graph = ont_graph  # type: Optional[GraphLike]
        self.data_graph_is_multigraph = isinstance(self.data_graph, (rdflib.Dataset, rdflib.ConjunctiveGraph))
        if self.ont_graph is not None and isinstance(self.ont_graph, (rdflib.Dataset, rdflib.ConjunctiveGraph)):
            self.ont_graph.default_union = True
        if self.ont_graph is not None and options['sparql_mode']:
            raise ReportableRuntimeError("Cannot use SPARQL Remote Graph Mode with extra Ontology Graph inoculation.")
        if shacl_graph is None:
            if options['sparql_mode']:
                raise ReportableRuntimeError(
                    "SHACL Shapes Graph must be a separate local graph or file when in SPARQL Remote Graph Mode."
                )
            shacl_graph = clone_graph(data_graph, identifier='shacl')
        assert isinstance(shacl_graph, rdflib.Graph), "shacl_graph must be a rdflib Graph object"
        self.shacl_graph = ShapesGraph(shacl_graph, self.debug, self.logger)  # type: ShapesGraph

        if options['use_js']:
            if options['sparql_mode']:
                raise ReportableRuntimeError("Cannot use SHACL-JS in SPARQL Remote Graph Mode.")
            is_js_installed = check_extra_installed('js')
            if is_js_installed:
                self.shacl_graph.enable_js()

    @classmethod
    def _load_default_options(cls, options_dict: dict):
        options_dict.setdefault('debug', False)
        options_dict.setdefault('advanced', False)
        options_dict.setdefault('inference', 'none')
        options_dict.setdefault('inplace', False)
        options_dict.setdefault('use_js', False)
        options_dict.setdefault('iterate_rules', False)
        options_dict.setdefault('abort_on_first', False)
        options_dict.setdefault('allow_infos', False)
        options_dict.setdefault('allow_warnings', False)
        options_dict.setdefault('sparql_mode', False)
        options_dict.setdefault('max_validation_depth', 15)
        options_dict.setdefault('focus_nodes', None)
        options_dict.setdefault('use_shapes', None)
        if 'logger' not in options_dict:
            options_dict['logger'] = logging.getLogger(__name__)
            if options_dict['debug']:
                options_dict['logger'].setLevel(logging.DEBUG)

    @classmethod
    def create_validation_report(cls, sg, conforms: bool, results: List[Tuple]):
        v_text = "Validation Report\nConforms: {}\n".format(str(conforms))
        result_len = len(results)
        if not conforms and result_len < 1:
            raise RuntimeError("A Non-Conformant Validation Report must have at least one result.")
        if result_len > 0:
            v_text += "Results ({}):\n".format(str(result_len))
        vg = rdflib.Graph(bind_namespaces='core')
        for p, n in sg.graph.namespace_manager.namespaces():
            vg.namespace_manager.bind(p, n)
        vr = BNode()
        vg.add((vr, RDF_type, SH_ValidationReport))
        vg.add((vr, SH_conforms, Literal(conforms)))
        cloned_nodes: Dict[Tuple[GraphLike, str], Union[BNode, URIRef]] = {}
        for result in iter(results):
            _d, _bn, _tr = result
            v_text += _d
            vg.add((vr, SH_result, _bn))
            for tr in iter(_tr):
                s, p, o = tr
                if isinstance(o, tuple):
                    source = o[0]
                    node = o[1]
                    if isinstance(node, Literal):
                        o = node  # No need to clone a literal from the data graph
                    else:
                        _id = str(node)
                        if (source, _id) in cloned_nodes:
                            o = cloned_nodes[(source, _id)]
                        elif isinstance(node, BNode):
                            cloned_nodes[(source, _id)] = o = clone_blank_node(source, node, vg, keepid=True)
                        else:
                            cloned_nodes[(source, _id)] = o = URIRef(_id)
                vg.add((s, p, o))
        return vg, v_text

    @property
    def target_graph(self) -> Union[GraphLike, None]:
        return self._target_graph

    def mix_in_ontology(self):
        if USE_FULL_MIXIN:
            if not self.data_graph_is_multigraph:
                return mix_graphs(self.data_graph, self.ont_graph, "inplace" if self.inplace else None)
            return mix_datasets(self.data_graph, self.ont_graph, "inplace" if self.inplace else None)
        if not self.data_graph_is_multigraph:
            if self.inplace:
                to_graph = self.data_graph
            else:
                to_graph = clone_graph(self.data_graph, identifier=self.data_graph.identifier)
            return inoculate(to_graph, self.ont_graph)
        return inoculate_dataset(
            self.data_graph,
            self.ont_graph,
            self.data_graph if self.inplace else None,
            URIRef("urn:pyshacl:inoculation"),
        )

    def make_executor(self) -> SHACLExecutor:
        return SHACLExecutor(
            validator=self,
            advanced_mode=bool(self.options.get('advanced', False)),
            abort_on_first=bool(self.options.get("abort_on_first", False)),
            allow_infos=bool(self.options.get("allow_infos", False)),
            allow_warnings=bool(self.options.get("allow_warnings", False)),
            iterate_rules=bool(self.options.get("iterate_rules", False)),
            sparql_mode=bool(self.options.get("sparql_mode", False)),
            max_validation_depth=self.options.get("max_validation_depth", 15),
            focus_nodes=self.options.get("focus_nodes", None),
            debug=self.debug,
        )

    def run(self):
        datagraph: Union[GraphLike, None] = self.target_graph
        if datagraph is not None:
            self._target_graph = datagraph
        else:
            has_cloned = False
            if self.ont_graph is not None:
                if self.inplace:
                    self.logger.debug("Adding ontology definitions to DataGraph")
                else:
                    self.logger.debug("Cloning DataGraph to temporary memory graph, to add ontology definitions.")
                # creates a copy of self.data_graph, doesn't modify it
                datagraph = self.mix_in_ontology()
                has_cloned = True
            else:
                datagraph = self.data_graph
            inference_option = self.options.get('inference', 'none')
            if self.inplace and self.debug:
                self.logger.debug("Skipping DataGraph clone because PySHACL is operating in inplace mode.")
            if inference_option and not self.pre_inferenced and str(inference_option) != "none":
                if self.options.get('sparql_mode', False):
                    raise ReportableRuntimeError("Cannot use any pre-inference option in SPARQL Remote Graph Mode.")
                if not has_cloned and not self.inplace:
                    self.logger.debug("Cloning DataGraph to temporary memory graph before pre-inferencing.")
                    datagraph = clone_graph(datagraph)
                    has_cloned = True
                self.logger.debug(f"Running pre-inferencing with option='{inference_option}'.")
                self._run_pre_inference(
                    datagraph, inference_option, URIRef("urn:pyshacl:inference"), logger=self.logger
                )
                self.pre_inferenced = True
            if not has_cloned and not self.inplace and self.options['advanced']:
                if self.options.get('sparql_mode', False):
                    raise ReportableRuntimeError("Cannot clone DataGraph in SPARQL Remote Graph Mode.")
                # We still need to clone in advanced mode, because of triple rules
                self.logger.debug("Forcing clone of DataGraph because advanced mode is enabled.")
                datagraph = clone_graph(datagraph)
                has_cloned = True
            if not has_cloned and not self.inplace:
                # No inferencing, no ont_graph, and no advanced mode, now implies inplace mode
                self.logger.debug("Running validation in-place, without modifying the DataGraph.")
                self.inplace = True
            self._target_graph = datagraph
        assert self._target_graph is not None
        if self.options.get("use_shapes", None) is not None and len(self.options["use_shapes"]) > 0:
            using_manually_specified_shapes = True
            expanded_use_shapes = []
            for s in self.options["use_shapes"]:
                s_lower = s.lower()
                if (
                    s_lower.startswith("http:")
                    or s_lower.startswith("https:")
                    or s_lower.startswith("urn:")
                    or s_lower.startswith("file:")
                ):
                    expanded_use_shapes.append(URIRef(s))
                else:
                    try:
                        expanded_use_shape = self.shacl_graph.graph.namespace_manager.expand_curie(s)
                    except ValueError:
                        expanded_use_shape = URIRef(s)
                    expanded_use_shapes.append(expanded_use_shape)
            shapes = self.shacl_graph.shapes_from_uris(expanded_use_shapes)
        else:
            using_manually_specified_shapes = False
            shapes = self.shacl_graph.shapes  # This property getter triggers shapes harvest.
        option_focus_nodes = self.options.get("focus_nodes", None)
        if option_focus_nodes is not None and len(option_focus_nodes) > 0:
            # Expand any CURIEs in the focus_nodes list
            expanded_focus_nodes: List[URIRef] = []
            for f in option_focus_nodes:
                f_lower = f.lower()
                if (
                    f_lower.startswith("http:")
                    or f_lower.startswith("https:")
                    or f_lower.startswith("urn:")
                    or f_lower.startswith("file:")
                ):
                    expanded_focus_nodes.append(URIRef(f))
                else:
                    try:
                        expanded_focus_node = self._target_graph.namespace_manager.expand_curie(f)
                    except ValueError:
                        expanded_focus_node = URIRef(f)
                    expanded_focus_nodes.append(expanded_focus_node)
            self.options["focus_nodes"] = expanded_focus_nodes
            specified_focus_nodes: Union[None, Sequence[URIRef]] = expanded_focus_nodes
        else:
            specified_focus_nodes = None
        executor = self.make_executor()

        # Special hack, if we are using manually specified shapes, and have
        # manually specified focus nodes, then we need to disable the
        # focus_nodes in the executor, because we apply the specified focus
        # nodes directly to the specified shapes.
        if using_manually_specified_shapes and specified_focus_nodes is not None:
            executor.focus_nodes = None

        if executor.advanced_mode:
            self.logger.debug("Activating SHACL-AF Features.")
            target_types = gather_target_types(self.shacl_graph)
            gather_from_shapes = None if not using_manually_specified_shapes else [s.node for s in shapes]
            advanced = {
                'functions': gather_functions(executor, self.shacl_graph),
                'rules': gather_rules(executor, self.shacl_graph, from_shapes=gather_from_shapes),
            }
            for s in shapes:
                s.set_advanced(True)
            apply_target_types(target_types)
        else:
            advanced = {}

        if specified_focus_nodes is not None and using_manually_specified_shapes:
            on_focus_nodes: Union[Sequence[URIRef], None] = specified_focus_nodes
        else:
            on_focus_nodes = None
        reports = []
        non_conformant = False
        if executor.abort_on_first and self.debug:
            self.logger.debug(
                "Abort on first error is enabled. Will exit at end of first Shape that fails validation."
            )

        if isinstance(self._target_graph, (rdflib.Dataset, rdflib.ConjunctiveGraph)):
            self._target_graph.default_union = True

        g = self._target_graph

        if self.debug:
            self.logger.debug(f"Validating DataGraph named {g.identifier}")
        if advanced:
            if advanced['functions']:
                apply_functions(executor, advanced['functions'], g)
            if advanced['rules']:
                if executor.sparql_mode:
                    self.logger.warning("Skipping SHACL Rules because operating in SPARQL Remote Graph Mode.")
                else:
                    apply_rules(executor, advanced['rules'], g, focus_nodes=on_focus_nodes)
        try:
            for s in shapes:
                _is_conform, _reports = s.validate(executor, g, focus=on_focus_nodes)
                non_conformant = non_conformant or (not _is_conform)
                reports.extend(_reports)
                if executor.abort_on_first and non_conformant:
                    break
        finally:
            if advanced and advanced['functions']:
                unapply_functions(advanced['functions'], g)
        v_report, v_text = self.create_validation_report(self.shacl_graph, not non_conformant, reports)
        return (not non_conformant), v_report, v_text


def assign_baked_in():
    if getattr(sys, 'frozen', False):
        # runs in a pyinstaller bundle
        HERE = sys._MEIPASS
    else:
        HERE = path.dirname(__file__)
    shacl_file = path.join(HERE, "assets", "shacl.pickle")
    add_baked_in("http://www.w3.org/ns/shacl", shacl_file)
    add_baked_in("https://www.w3.org/ns/shacl", shacl_file)
    add_baked_in("http://www.w3.org/ns/shacl.ttl", shacl_file)
    shacl_shacl_file = path.join(HERE, "assets", "shacl-shacl.pickle")
    add_baked_in("http://www.w3.org/ns/shacl-shacl", shacl_shacl_file)
    add_baked_in("https://www.w3.org/ns/shacl-shacl", shacl_shacl_file)
    add_baked_in("http://www.w3.org/ns/shacl-shacl.ttl", shacl_shacl_file)
    schema_file = path.join(HERE, "assets", "schema.pickle")
    add_baked_in("http://datashapes.org/schema", schema_file)
    add_baked_in("https://datashapes.org/schema", schema_file)
    add_baked_in("http://datashapes.org/schema.ttl", schema_file)
    dash_file = path.join(HERE, "assets", "dash.pickle")
    add_baked_in("http://datashapes.org/dash", dash_file)
    add_baked_in("https://datashapes.org/dash", dash_file)
    add_baked_in("http://datashapes.org/dash.ttl", dash_file)



# -*- coding: utf-8 -*-
#
"""
https://www.w3.org/TR/shacl/#core-components-value-type
"""

import abc
import re
import typing
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Set, Tuple

from rdflib import BNode, Literal, URIRef

from pyshacl.consts import (
    SH,
    RDF_type,
    SH_ask,
    SH_focusNode,
    SH_jsFunctionName,
    SH_NodeConstraintComponent,
    SH_parameter,
    SH_path,
    SH_PropertyConstraintComponent,
    SH_resultMessage,
    SH_resultPath,
    SH_resultSeverity,
    SH_select,
    SH_sourceConstraint,
    SH_sourceConstraintComponent,
    SH_sourceShape,
    SH_ValidationResult,
    SH_value,
    SH_Violation,
)
from pyshacl.errors import ConstraintLoadError
from pyshacl.parameter import SHACLParameter
from pyshacl.pytypes import GraphLike, SHACLExecutor
from pyshacl.rdfutil import stringify_node

if TYPE_CHECKING:
    from pyshacl.pytypes import RDFNode
    from pyshacl.shape import Shape
    from pyshacl.shapes_graph import ShapesGraph


class ConstraintComponent(object, metaclass=abc.ABCMeta):
    __slots__ = ('shape',)

    """
    Abstract Constraint Component Class
    All Constraint Components must inherit from this class.
    """

    # True if constraint component is defined as "shape-expecting"
    shape_expecting: bool = False

    # True if constraint component is defined as "list-taking"
    list_taking: bool = False

    shacl_constraint_component: URIRef = URIRef("urn:notimplemented")

    def __init__(self, shape: 'Shape') -> None:
        """

        :param shape:
        :type shape: Shape
        """
        self.shape = shape  # type: Shape

    @classmethod
    @abc.abstractmethod
    def constraint_parameters(cls) -> List[URIRef]:
        raise NotImplementedError()  # pragma: no cover

    @classmethod
    @abc.abstractmethod
    def constraint_name(cls) -> str:
        raise NotImplementedError()  # pragma: no cover

    @abc.abstractmethod
    def evaluate(
        self, executor: SHACLExecutor, target_graph: GraphLike, focus_value_nodes: Dict, _evaluation_path: List
    ):
        raise NotImplementedError()  # pragma: no cover

    def make_generic_messages(self, datagraph: GraphLike, focus_node, value_node) -> List[Literal]:
        return []

    def __str__(self) -> str:
        c_name = str(self.__class__.__name__)
        shape_id = str(self.shape)
        return "<{} on {}>".format(c_name, shape_id)

    def recursion_triggers(self, _evaluation_path, trigger_depth=3) -> Optional[List['RDFNode']]:
        shape = self.shape
        eval_length = len(_evaluation_path)
        if eval_length < 4:
            return None
        maybe_recursive = []
        _shape, _self = _evaluation_path[eval_length - 2 :]
        if _shape is not shape or _self is not self:
            raise RuntimeError("Bad evaluation path construction")
        prev_shape, prev_constraint = _evaluation_path[eval_length - 4 : eval_length - 2]
        lookback_len = trigger_depth * 2
        if isinstance(prev_constraint, ConstraintComponent):
            if (
                self.shacl_constraint_component is SH_PropertyConstraintComponent
                and prev_constraint.shacl_constraint_component is SH_NodeConstraintComponent
            ) or (
                self.shacl_constraint_component is SH_NodeConstraintComponent
                and prev_constraint.shacl_constraint_component is SH_PropertyConstraintComponent
            ):
                lookback_len = trigger_depth * 4
        if eval_length < lookback_len:
            return None
        seen_before = [i for i, x in enumerate(_evaluation_path[: eval_length - 2]) if x is shape]
        for s in seen_before:
            for i, p in enumerate(_evaluation_path[s + 1 : -2]):
                if isinstance(p, ConstraintComponent):
                    if p.shape is shape and p.__class__ == self.__class__:
                        try:
                            next_shape = _evaluation_path[s + 1 + i + 1]
                            maybe_recursive.append(next_shape)
                        except IndexError:
                            pass
                    break
        return maybe_recursive

    def make_v_result_description(
        self,
        datagraph: GraphLike,
        focus_node: 'RDFNode',
        severity: URIRef,
        value_node: Optional['RDFNode'],
        messages: List[str],
        result_path=None,
        constraint_component=None,
        source_constraint=None,
        extra_messages: Optional[Iterable] = None,
        bound_vars=None,
    ):
        """
        :param datagraph:
        :type datagraph: rdflib.Graph | rdflib.ConjunctiveGraph | rdflib.Dataset
        :param focus_node:
        :type focus_node: RDFNode
        :param severity:
        :type value_node: rdflib.URIRef
        :param value_node:
        :type value_node: rdflib.term.Identifier | None
        :param messages:
        :type messages: List[str]
        :param result_path:
        :param bound_vars:
        :param constraint_component:
        :param source_constraint:
        :param extra_messages:
        :type extra_messages: collections.abc.Iterable | None
        :return:
        """
        sg = self.shape.sg.graph
        constraint_component = constraint_component or self.shacl_constraint_component
        constraint_name = self.constraint_name()
        if severity == SH_Violation:
            severity_desc = "Constraint Violation"
        else:
            severity_desc = "Validation Result"
        source_shape_text = stringify_node(sg, self.shape.node)
        severity_node_text = stringify_node(sg, severity)
        try:
            focus_node_text = stringify_node(datagraph or sg, focus_node)
        except (LookupError, ValueError):
            # focus node doesn't exist in the datagraph. We can deal.
            focus_node_text = str(focus_node)
        desc = "{} in {} ({}):\n\tSeverity: {}\n\tSource Shape: {}\n\tFocus Node: {}\n".format(
            severity_desc,
            constraint_name,
            str(constraint_component),
            severity_node_text,
            source_shape_text,
            focus_node_text,
        )
        if value_node is not None:
            try:
                val_node_string = stringify_node(datagraph or sg, value_node)
            except (LookupError, ValueError):
                # value node doesn't exist in the datagraph.
                val_node_string = str(value_node)
            desc += "\tValue Node: {}\n".format(val_node_string)
        if result_path is None and self.shape.is_property_shape:
            result_path = self.shape.path()
        if result_path:
            result_path_text = stringify_node(sg, result_path)
            desc += "\tResult Path: {}\n".format(result_path_text)
        if source_constraint:
            sc_text = stringify_node(sg, source_constraint)
            desc += "\tSource Constraint: {}\n".format(sc_text)
        if extra_messages:
            for m in iter(extra_messages):
                if m in messages:
                    continue
                if isinstance(m, Literal):
                    msg = str(m.value)
                    if bound_vars is not None:
                        msg = self._format_sparql_based_result_message(msg, bound_vars)
                    desc += "\tMessage: {}\n".format(msg)
                else:  # pragma: no cover
                    desc += "\tMessage: {}\n".format(str(m))
        for m in messages:
            if isinstance(m, Literal):
                msg = str(m.value)
                if bound_vars is not None:
                    msg = self._format_sparql_based_result_message(msg, bound_vars)
                desc += "\tMessage: {}\n".format(msg)
            else:  # pragma: no cover
                desc += "\tMessage: {}\n".format(str(m))
        return desc

    def make_v_result(
        self,
        datagraph: GraphLike,
        focus_node: 'RDFNode',
        value_node: Optional['RDFNode'] = None,
        result_path: Optional['RDFNode'] = None,
        constraint_component: Optional['RDFNode'] = None,
        source_constraint: Optional['RDFNode'] = None,
        extra_messages: Optional[Iterable] = None,
        bound_vars=None,
    ):
        """
        :param datagraph:
        :type datagraph: rdflib.Graph | rdflib.ConjunctiveGraph | rdflib.Dataset
        :param focus_node:
        :type focus_node: RDFNode
        :param value_node:
        :type value_node: RDFNode | None
        :param result_path:
        :type result_path: RDFNode | None
        :param constraint_component:
        :param source_constraint:
        :param extra_messages:
        :type extra_messages: collections.abc.Iterable | None
        :param bound_vars:
        :return:
        """
        constraint_component = constraint_component or self.shacl_constraint_component
        severity = self.shape.severity
        sg = self.shape.sg.graph
        r_triples: List[Tuple[RDFNode, RDFNode, Any]] = list()
        r_node = BNode()
        r_triples.append((r_node, RDF_type, SH_ValidationResult))
        r_triples.append((r_node, SH_sourceConstraintComponent, (sg, constraint_component)))
        r_triples.append((r_node, SH_sourceShape, (sg, self.shape.node)))
        r_triples.append((r_node, SH_resultSeverity, severity))
        r_triples.append((r_node, SH_focusNode, (datagraph or sg, focus_node)))
        if value_node is not None:
            r_triples.append((r_node, SH_value, (datagraph, value_node)))
        if result_path is None and self.shape.is_property_shape:
            result_path = self.shape.path()
        if result_path is not None:
            r_triples.append((r_node, SH_resultPath, (sg, result_path)))
        if source_constraint is not None:
            r_triples.append((r_node, SH_sourceConstraint, (sg, source_constraint)))
        messages = list(self.shape.message)
        if extra_messages:
            for m in iter(extra_messages):
                if m in messages:
                    continue
                if isinstance(m, Literal):
                    msg = str(m.value)
                    if bound_vars is not None:
                        msg = self._format_sparql_based_result_message(msg, bound_vars)
                        m = Literal(msg)
                r_triples.append((r_node, SH_resultMessage, m))
        elif not messages:
            messages = self.make_generic_messages(datagraph, focus_node, value_node) or messages
        for m in messages:
            if isinstance(m, Literal):
                msg = str(m.value)
                if bound_vars is not None:
                    msg = self._format_sparql_based_result_message(msg, bound_vars)
                    m = Literal(msg)
            r_triples.append((r_node, SH_resultMessage, m))
        desc = self.make_v_result_description(
            datagraph,
            focus_node,
            severity,
            value_node,
            messages,
            result_path=result_path,
            constraint_component=constraint_component,
            source_constraint=source_constraint,
            extra_messages=extra_messages,
            bound_vars=bound_vars,
        )
        return desc, r_node, r_triples

    def _format_sparql_based_result_message(self, msg, bound_vars):
        if bound_vars is None:
            return msg
        fdict = {}
        if isinstance(bound_vars, (tuple, list)):
            if len(bound_vars) == 4:
                fdict.update(bound_vars[3])
                bound_vars = bound_vars[:3]
            if len(bound_vars) == 3:
                if bound_vars[0] is not None:
                    fdict['this'] = bound_vars[0]
                if bound_vars[1] is not None:
                    fdict['path'] = bound_vars[1]
                if bound_vars[2] is not None:
                    fdict['value'] = bound_vars[2]

        elif isinstance(bound_vars, dict):
            fdict.update(bound_vars)
        else:
            return msg
        for var, val in fdict.items():
            substring = "{{[?$]{}}}".format(var)
            msg = re.sub(substring, str(val), msg)
        return msg


SH_nodeValidator = SH.nodeValidator
SH_propertyValidator = SH.propertyValidator
SH_validator = SH.validator
SH_optional = SH.optional
SH_SPARQLSelectValidator = SH.SPARQLSelectValidator
SH_SPARQLAskValidator = SH.SPARQLAskValidator
SH_JSValidator = SH.JSValidator


class CustomConstraintComponentFactory(object):
    __slots__: Tuple = tuple()

    def __new__(cls, shacl_graph: 'ShapesGraph', node):
        self: List[Any] = list()
        self.append(shacl_graph)
        self.append(node)
        optional_params = []
        mandatory_params = []
        param_nodes = set(shacl_graph.objects(node, SH_parameter))
        if len(param_nodes) < 1:
            # TODO:coverage: we don't have any tests for invalid constraints
            raise ConstraintLoadError(
                "A sh:ConstraintComponent must have at least one value for sh:parameter",
                "https://www.w3.org/TR/shacl/#constraint-components-parameters",
            )
        for param_node in iter(param_nodes):
            path_nodes = set(shacl_graph.objects(param_node, SH_path))
            if len(path_nodes) < 1:
                # TODO:coverage: we don't have any tests for invalid constraints
                raise ConstraintLoadError(
                    "A sh:ConstraintComponent parameter value must have at least one value for sh:path",
                    "https://www.w3.org/TR/shacl/#constraint-components-parameters",
                )
            elif len(path_nodes) > 1:
                # TODO:coverage: we don't have any tests for invalid constraints
                raise ConstraintLoadError(
                    "A sh:ConstraintComponent parameter value must have at most one value for sh:path",
                    "https://www.w3.org/TR/shacl/#constraint-components-parameters",
                )
            path = next(iter(path_nodes))
            parameter = SHACLParameter(shacl_graph, param_node, path=path, logger=None)  # pass in logger?
            if parameter.optional:
                optional_params.append(parameter)
            else:
                mandatory_params.append(parameter)
        if len(mandatory_params) < 1:
            # TODO:coverage: we don't have any tests for invalid constraint components
            raise ConstraintLoadError(
                "A sh:ConstraintComponent must have at least one non-optional parameter.",
                "https://www.w3.org/TR/shacl/#constraint-components-parameters",
            )
        self.append(mandatory_params + optional_params)

        validator_node_set = set(shacl_graph.graph.objects(node, SH_validator))
        node_val_node_set = set(shacl_graph.graph.objects(node, SH_nodeValidator))
        prop_val_node_set = set(shacl_graph.graph.objects(node, SH_propertyValidator))
        validator_node_set = validator_node_set.difference(node_val_node_set)
        validator_node_set = validator_node_set.difference(prop_val_node_set)
        self.append(validator_node_set)
        self.append(node_val_node_set)
        self.append(prop_val_node_set)
        is_sparql_constraint_component = False
        is_js_constraint_component = False
        for s in (validator_node_set, node_val_node_set, prop_val_node_set):
            for v in s:
                v_types = set(shacl_graph.graph.objects(v, RDF_type))
                if SH_SPARQLAskValidator in v_types or SH_SPARQLSelectValidator in v_types:
                    is_sparql_constraint_component = True
                    break
                elif SH_JSValidator in v_types:
                    is_js_constraint_component = True
                    break
                v_props = set(p[0] for p in shacl_graph.graph.predicate_objects(v))
                if SH_ask in v_props or SH_select in v_props:
                    is_sparql_constraint_component = True
                    break
                elif SH_jsFunctionName in v_props:
                    is_js_constraint_component = True
                    break
                if is_sparql_constraint_component:
                    raise ConstraintLoadError(
                        "Found a mix of SPARQL-based validators and non-SPARQL validators on a SPARQLConstraintComponent.",  # noqa
                        'https://www.w3.org/TR/shacl/#constraint-components-validators',
                    )
                elif is_js_constraint_component:
                    raise ConstraintLoadError(
                        "Found a mix of JS-based validators and non-JS validators on a JSConstraintComponent.",
                        'https://www.w3.org/TR/shacl/#constraint-components-validators',
                    )
        if is_sparql_constraint_component:
            from pyshacl.constraints.sparql.sparql_based_constraint_components import SPARQLConstraintComponent

            return SPARQLConstraintComponent(*self)
        elif is_js_constraint_component and shacl_graph.js_enabled:
            from pyshacl.extras.js.constraint_component import JSConstraintComponent

            return JSConstraintComponent(*self)
        else:
            return CustomConstraintComponent(*self)


class CustomConstraintComponent(object):
    __slots__: Tuple = ('sg', 'node', 'parameters', 'validators', 'node_validators', 'property_validators')

    if typing.TYPE_CHECKING:
        sg: ShapesGraph
        node: Any
        parameters: List[SHACLParameter]
        validators: Set
        node_validators: Set
        property_validators: Set

    def __new__(cls, shacl_graph: 'ShapesGraph', node, parameters, validators, node_validators, property_validators):
        self = super(CustomConstraintComponent, cls).__new__(cls)
        self.sg = shacl_graph
        self.node = node
        self.parameters = parameters
        self.validators = validators
        self.node_validators = node_validators
        self.property_validators = property_validators
        return self

    def make_validator_for_shape(self, shape: 'Shape'):
        raise ConstraintLoadError(
            "A Custom Constraint must include one of a SPARQLConstraintComponent validator or a JSConstraint validator.",
            "https://www.w3.org/TR/shacl/#constraint-components-validators",
        )



#
#
from typing import TYPE_CHECKING, List, Optional, Sequence, Union

import rdflib

from pyshacl.consts import SH
from pyshacl.errors import ReportableRuntimeError
from pyshacl.rules.shacl_rule import SHACLRule

from .js_executable import JSExecutable

if TYPE_CHECKING:
    from rdflib.term import URIRef

    from pyshacl.pytypes import GraphLike, RDFNode, SHACLExecutor
    from pyshacl.shape import Shape
    from pyshacl.shapes_graph import ShapesGraph

SH_JSRule = SH.JSRule

JS_RULE_ITERATE_LIMIT = 100


class JSRule(SHACLRule):
    __slots__ = ('js_exe',)

    def __init__(self, executor: 'SHACLExecutor', shape: 'Shape', rule_node, **kwargs):
        super(JSRule, self).__init__(executor, shape, rule_node, **kwargs)
        shapes_graph: 'ShapesGraph' = shape.sg
        self.js_exe = JSExecutable(shapes_graph, rule_node)

    def apply(
        self,
        data_graph: 'GraphLike',
        focus_nodes: Union[Sequence['RDFNode'], None] = None,
        target_graph_identifier: Optional['URIRef'] = None,
    ) -> int:
        focus_list: Sequence['RDFNode']
        if focus_nodes is not None:
            focus_list = list(focus_nodes)
        else:
            focus_list = list(self.shape.focus_nodes(data_graph))
        if self.executor.focus_nodes is not None and len(self.executor.focus_nodes) > 0:
            filtered_focus_nodes: List[Union[rdflib.URIRef]] = []
            for _fo in focus_list:  # type: RDFNode
                if isinstance(_fo, rdflib.URIRef) and _fo in self.executor.focus_nodes:
                    filtered_focus_nodes.append(_fo)
            len_filtered_focus = len(filtered_focus_nodes)
            if len_filtered_focus < 1:
                return 0
            focus_list = filtered_focus_nodes
        all_added = 0
        iterate_limit = int(JS_RULE_ITERATE_LIMIT)
        while True:
            if iterate_limit < 1:
                raise ReportableRuntimeError(f"JS rule iteration exceeded iteration limit of {JS_RULE_ITERATE_LIMIT}.")
            iterate_limit -= 1
            added = 0
            applicable_nodes = self.filter_conditions(focus_list, data_graph)
            sets_to_add = []
            for a in applicable_nodes:
                args_map = {"this": a}
                results = self.js_exe.execute(data_graph, args_map, mode="construct")
                triples = results['_result']
                this_added = False
                if triples is not None and isinstance(triples, (list, tuple)):
                    set_to_add = set()
                    for t in triples:
                        s, p, o = tr = t[:3]
                        if not this_added and tr not in data_graph:
                            this_added = True
                        set_to_add.add(tr)
                    sets_to_add.append(set_to_add)
                if this_added:
                    added += 1
            if added > 0:
                if isinstance(data_graph, (rdflib.Dataset, rdflib.ConjunctiveGraph)):
                    if target_graph_identifier is not None:
                        target_graph = data_graph.get_context(target_graph_identifier)
                    else:
                        target_graph = data_graph.default_context
                else:
                    target_graph = data_graph
                for s in sets_to_add:
                    for t in s:
                        target_graph.add(t)
                all_added += added
                if self.iterate:
                    continue  # Jump up to iterate
                else:
                    break  # Don't iterate
            break
        return all_added



import logging
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Tuple, Union

import rdflib
from rdflib import BNode, Literal, URIRef
from rdflib.util import from_n3

from pyshacl.consts import (
    RDF_object,
    RDF_predicate,
    RDF_subject,
    RDF_type,
    RDFS_Resource,
    SH_conforms,
    SH_detail,
    SH_focusNode,
    SH_result,
    SH_resultMessage,
    SH_resultPath,
    SH_ValidationReport,
    SH_value,
)
from pyshacl.errors import ReportableRuntimeError, ValidationFailure
from pyshacl.functions import apply_functions, gather_functions
from pyshacl.pytypes import GraphLike, RDFNode
from pyshacl.rdfutil import compare_blank_node, compare_node, order_graph_literal, stringify_node

if TYPE_CHECKING:
    from pyshacl.validator import Validator


def clean_validation_reports(actual_graph, actual_report, expected_graph, expected_report):
    # remove rdfs-added stuff
    # remove resultMessage if expected_report does not include result_message
    # remove sh:detail if expected_report does not include details
    # expected_graph.remove((expected_report, RDF_type, RDFS_Resource))
    # actual_graph.remove((actual_report, RDF_type, RDFS_Resource))
    expected_graph.remove((None, RDF_type, RDFS_Resource))
    actual_graph.remove((None, RDF_type, RDFS_Resource))
    expected_results = list(expected_graph.objects(expected_report, SH_result))
    actual_results = list(actual_graph.objects(actual_report, SH_result))
    er_has_messages = None
    er_has_details = False
    for er in expected_results:
        expected_graph.remove((er, RDF_type, RDFS_Resource))
        er_has_messages = list(expected_graph.objects(er, SH_resultMessage))
        er_has_details = er_has_details or expected_graph.value(er, SH_detail) is not None
        # sourceShapes = list(expected_graph.objects(er, SH_sourceShape))
        # for s in sourceShapes:
        #     expected_graph.remove((s, RDF_type, RDFS_Resource))
        # resultPaths = list(expected_graph.objects(er, SH_resultPath))
        # for r in resultPaths:
        #     expected_graph.remove((r, RDF_type, RDFS_Resource))
        # sourceConstraints = list(expected_graph.objects(er, SH_sourceConstraint))
        # for s in sourceConstraints:
        #     expected_graph.remove((s, RDF_type, RDFS_Resource))
    if er_has_messages and len(er_has_messages) > 0:
        # keep messages in actual
        pass
    else:
        for ar in actual_results:
            actual_graph.remove((ar, SH_resultMessage, None))
    if not er_has_details:
        # If no expected result had details, remove all details from actual
        for ar in actual_results:
            for detail in actual_graph.objects(ar, SH_detail):
                actual_graph -= actual_graph.cbd(detail)
                actual_graph.remove((ar, SH_detail, detail))
    return True


def compare_validation_reports(
    report_graph: GraphLike, expected_graph: GraphLike, expected_result, log: logging.Logger
):
    expected_conforms_i = expected_graph.objects(expected_result, SH_conforms)
    expected_conforms = set(expected_conforms_i)
    if len(expected_conforms) < 1:  # pragma: no cover
        raise ReportableRuntimeError(
            "Cannot check the expected result, the given expectedResult does not have an sh:conforms."
        )
    expected_conform = next(iter(expected_conforms))
    expected_result_nodes_i = expected_graph.objects(expected_result, SH_result)
    expected_result_nodes = list(expected_result_nodes_i)
    expected_result_node_count = len(expected_result_nodes)

    validation_reports = report_graph.subjects(RDF_type, SH_ValidationReport)
    validation_reports_set = set(validation_reports)
    if len(validation_reports_set) < 1:  # pragma: no cover
        raise ReportableRuntimeError(
            "Cannot check the validation report, the report graph does not contain a ValidationReport"
        )
    validation_report = next(iter(validation_reports_set))
    clean_validation_reports(report_graph, validation_report, expected_graph, expected_result)
    eq = compare_blank_node(report_graph, validation_report, expected_graph, expected_result)
    if eq != 0:
        return False
    report_conforms_i = report_graph.objects(validation_report, SH_conforms)
    report_conforms = set(report_conforms_i)
    if len(report_conforms) < 1:  # pragma: no cover
        raise ReportableRuntimeError(
            "Cannot check the validation report, the report graph does not have an sh:conforms."
        )
    report_conform = next(iter(report_conforms))

    if (
        isinstance(expected_conform, Literal)
        and isinstance(report_conform, Literal)
        and bool(expected_conform.value) != bool(report_conform.value)
    ):
        # TODO:coverage: write a test for this
        log.error("Expected Result Conforms value is different from Validation Report's Conforms value.")
        return False

    report_result_nodes_i = report_graph.objects(validation_report, SH_result)
    report_result_nodes = list(report_result_nodes_i)
    report_result_node_count = len(report_result_nodes)

    if expected_result_node_count != report_result_node_count:
        # TODO:coverage: write a test for this
        log.error(
            "Number of expected result's sh:result entries is different from Validation Report's sh:result entries.\n"
            "Expected {}, got {}.".format(expected_result_node_count, report_result_node_count)
        )
        return False

    expected_results_dict: Dict[Tuple[str, str, str], Any] = {}
    report_results_dict: Dict[Tuple[str, str, str], Any] = {}
    for result_nodes, result_graph, dest_dict in (
        (expected_result_nodes, expected_graph, expected_results_dict),
        (report_result_nodes, report_graph, report_results_dict),
    ):
        for result in result_nodes:
            result_focus_i = result_graph.objects(result, SH_focusNode)
            result_focus_list = list(result_focus_i)
            if len(result_focus_list) > 0:
                f_node = result_focus_list[0]
                if isinstance(f_node, Literal):
                    result_focus = str(f_node)
                elif isinstance(f_node, BNode):
                    # result_value = "_:" + str(v_node)
                    # Can't compare BNodes because they are
                    # different in the shapes graph than the data graph
                    result_focus = "BNode"
                else:
                    result_focus = stringify_node(result_graph, f_node)
            else:
                result_focus = ""
            result_value_i = result_graph.objects(result, SH_value)
            result_value_list = list(result_value_i)
            if len(result_value_list) > 0:
                v_node = result_value_list[0]
                if isinstance(v_node, Literal):
                    result_value = str(v_node)
                elif isinstance(v_node, BNode):
                    # result_value = "_:" + str(v_node)
                    # Can't compare BNodes because they are
                    # different in the shapes graph than the data graph
                    result_value = "BNode"
                else:
                    result_value = stringify_node(result_graph, v_node)
            else:
                result_value = ""
            result_path_i = result_graph.objects(result, SH_resultPath)
            result_path_list = list(result_path_i)
            if len(result_path_list) > 0:
                result_path = stringify_node(result_graph, result_path_list[0])
            else:
                result_path = ""
            dest_dict[(result_focus, result_value, result_path)] = result
    not_found_results = 0
    for expected_focus, expected_value, expected_path in expected_results_dict.keys():
        if (expected_focus, expected_value, expected_path) not in report_results_dict:
            log.error(
                "Expected result not found in Validation Report.\nExpected focus: {}, value: {}, path: {}.".format(
                    expected_focus, expected_value, expected_path
                )
            )
            not_found_results += 1
    if not_found_results > 0:
        return False
    return True


def compare_inferencing_reports(data_graph: GraphLike, expected_graph: GraphLike, expected_results: Union[List, Set]):
    all_good = True
    for expected_result in expected_results:
        expected_objects = set(expected_graph.objects(expected_result, RDF_object))
        if len(expected_objects) < 1:
            raise ReportableRuntimeError(
                "Cannot check the expected result, the given expectedResult does not have an rdf:object."
            )
        expected_object = next(iter(expected_objects))
        expected_subjects = set(expected_graph.objects(expected_result, RDF_subject))
        if len(expected_subjects) < 1:
            raise ReportableRuntimeError(
                "Cannot check the expected result, the given expectedResult does not have an rdf:subject."
            )
        expected_subject = next(iter(expected_subjects))
        expected_predicates = set(expected_graph.objects(expected_result, RDF_predicate))
        if len(expected_predicates) < 1:
            raise ReportableRuntimeError(
                "Cannot check the expected result, the given expectedResult does not have an rdf:predicate."
            )
        expected_predicate = next(iter(expected_predicates))
        if isinstance(expected_object, Literal):
            found_objs = set(data_graph.objects(expected_subject, expected_predicate))
            if len(found_objs) < 1:
                all_good = False
                print("Found no sub/pred matching {} {}".format(expected_subject, expected_predicate))
                continue
            found = False
            for o in found_objs:
                if isinstance(o, Literal):
                    found = 0 == order_graph_literal(expected_graph, expected_object, data_graph, o)
                    if found:
                        break
            if not found:
                print(
                    "Found no sub/pred/obj matching {} {} {}".format(
                        expected_subject, expected_predicate, expected_object
                    )
                )
            all_good = all_good and found
            continue

        elif isinstance(expected_object, BNode):
            found_objs = set(data_graph.objects(expected_subject, expected_predicate))
            if len(found_objs) < 1:
                all_good = False
                print("Found no sub/pred matching {} {}".format(expected_subject, expected_predicate))
                continue
            found = False
            for o in found_objs:
                if isinstance(o, BNode):
                    found = 0 == compare_blank_node(expected_graph, expected_object, data_graph, o)
                    if found:
                        break
            if not found:
                print(
                    "Found no sub/pred/obj matching {} {} {}".format(
                        expected_subject, expected_predicate, expected_object
                    )
                )
            all_good = all_good and found
            continue
        else:
            found_triples = set(data_graph.triples((expected_subject, expected_predicate, expected_object)))
            if len(found_triples) < 1:
                all_good = False

    return all_good


def check_dash_result(
    validator: 'Validator',
    report_graph: GraphLike,
    expected_result_graph: GraphLike,
    log: Union[logging.Logger, None] = None,
):
    DASH = rdflib.namespace.Namespace('http://datashapes.org/dash#')
    DASH_GraphValidationTestCase = DASH.GraphValidationTestCase
    DASH_InferencingTestCase = DASH.InferencingTestCase
    DASH_FunctionTestCase = DASH.FunctionTestCase
    DASH_expectedResult = DASH.expectedResult
    DASH_expression = DASH.expression
    was_default_union = None
    if log is None:
        log = logging.getLogger(__name__)
    if isinstance(expected_result_graph, (rdflib.ConjunctiveGraph, rdflib.Dataset)):
        was_default_union = expected_result_graph.default_union
        expected_result_graph.default_union = True  # Force default-union to make all of this a bit easier
    gv_test_cases = expected_result_graph.subjects(RDF_type, DASH_GraphValidationTestCase)
    gv_test_cases_set = set(gv_test_cases)
    inf_test_cases = expected_result_graph.subjects(RDF_type, DASH_InferencingTestCase)
    inf_test_cases_set = set(inf_test_cases)
    fn_test_cases = expected_result_graph.subjects(RDF_type, DASH_FunctionTestCase)
    fn_test_cases_set = set(fn_test_cases)
    if len(gv_test_cases_set) > 0:
        test_case = next(iter(gv_test_cases_set))
        expected_results = expected_result_graph.objects(test_case, DASH_expectedResult)
        expected_results_set = set(expected_results)
        if len(expected_results_set) < 1:  # pragma: no cover
            raise ReportableRuntimeError(
                "Cannot check the expected result, the given GraphValidationTestCase does not have an expectedResult."
            )
        expected_result = next(iter(expected_results_set))
        gv_res: Union[bool, None] = compare_validation_reports(
            report_graph, expected_result_graph, expected_result, log
        )
    else:
        gv_res = None
    if len(inf_test_cases_set) > 0:
        data_graph = validator.target_graph
        assert data_graph is not None
        if isinstance(data_graph, (rdflib.ConjunctiveGraph, rdflib.Dataset)):
            named_graphs = list(data_graph.contexts())
        else:
            named_graphs = [data_graph]
        inf_res: Union[bool, None] = True
        for test_case in inf_test_cases_set:
            expected_results = expected_result_graph.objects(test_case, DASH_expectedResult)
            expected_results_set = set(expected_results)
            if len(expected_results_set) < 1:  # pragma: no cover
                raise ReportableRuntimeError(
                    "Cannot check the expected result, the given InferencingTestCase does not have an expectedResult."
                )
            found = False
            for g in named_graphs:
                found = found or compare_inferencing_reports(g, expected_result_graph, expected_results_set)
            inf_res = inf_res and found
    else:
        inf_res = None
    if len(fn_test_cases_set) > 0:
        executor = validator.make_executor()
        data_graph = validator.target_graph
        assert data_graph is not None
        fns = gather_functions(executor, validator.shacl_graph)
        apply_functions(executor, fns, data_graph)
        fn_res: Union[bool, None] = True
        for test_case in fn_test_cases_set:
            expected_results_set = set(expected_result_graph.objects(test_case, DASH_expectedResult))
            if len(expected_results_set) < 1:  # pragma: no cover
                raise ReportableRuntimeError(
                    "Cannot check the expected result, the given FunctionTestCase does not have an expectedResult."
                )
            expected_result = next(iter(expected_results_set))
            expressions = set(expected_result_graph.objects(test_case, DASH_expression))
            if len(expressions) < 1:
                raise ReportableRuntimeError(
                    "Cannot check the expected result, the given FunctionTestCase does not have an expression."
                )
            expression_node = next(iter(expressions))
            expression = str(expression_node).strip()
            parts = [e.strip() for e in expression.split("(", 1)]
            eargs: List[Optional[Union[str, RDFNode]]]
            if len(parts) < 1:
                expression = parts[0]
                eargs = []
            else:
                expression, sargs = parts
                sargs = sargs.rstrip(")")
                if len(sargs) < 1:
                    eargs_str_list: List[str] = []
                else:
                    eargs_str_list = [a.strip() for a in sargs.split(',')]
                eargs = []
                for e in eargs_str_list:
                    eargs.append(
                        from_n3(e, None, expected_result_graph.store, expected_result_graph.namespace_manager)  # type: ignore[arg-type]
                    )
            find_uri = from_n3(expression, None, expected_result_graph.store, expected_result_graph.namespace_manager)  # type: ignore[arg-type]
            if find_uri is None or not isinstance(find_uri, (str, URIRef, Literal, BNode)):
                raise ReportableRuntimeError("Cannot execute function {}.\nBad declaration format.".format(find_uri))
            try:
                fn, options = validator.shacl_graph.get_shacl_function(find_uri)
            except KeyError:
                raise ReportableRuntimeError(
                    "Cannot execute function {}.\nCannot find it in the ShapesGraph object.".format(find_uri)
                )
            result = fn(data_graph, *eargs)
            fn_res = fn_res and 0 == compare_node(expected_result_graph, expected_result, data_graph, result)

    else:
        fn_res = None
    if was_default_union is not None:
        expected_result_graph.default_union = was_default_union
    if gv_res is None and inf_res is None and fn_res is None:  # pragma: no cover
        raise ReportableRuntimeError(
            "Cannot check the expected result, the given expected result graph does not have a GraphValidationTestCase or InferencingTestCase."
        )
    return (gv_res or gv_res is None) and (inf_res or inf_res is None) and (fn_res or fn_res is None)


def check_sht_result(
    report_graph: GraphLike,
    sht_graph: GraphLike,
    sht_result_node: Union[URIRef, BNode],
    log: Union[logging.Logger, None] = None,
):
    SHT = rdflib.namespace.Namespace('http://www.w3.org/ns/shacl-test#')
    types = set(sht_graph.objects(sht_result_node, RDF_type))
    expected_failure = sht_result_node == SHT.Failure
    if log is None:
        log = logging.getLogger(__name__)
    if expected_failure and isinstance(report_graph, ValidationFailure):
        return True
    elif isinstance(report_graph, ValidationFailure):
        # TODO:coverage: write a test for this
        log.error("Validation Report indicates a Validation Failure, but the SHT entry does not expect a failure.")
        return False
    elif expected_failure:
        # TODO:coverage: write a test for this
        log.error(
            "SHT expects a Validation Failure, but the Validation Report does not indicate a Validation Failure."
        )
        return False
    if SH_ValidationReport not in types:
        raise ReportableRuntimeError("SHT expected result must have type sh:ValidationReport")
    return compare_validation_reports(report_graph, sht_graph, sht_result_node, log)



# -*- coding: latin-1 -*-
#
import os
import sys

from pyshacl.cli import main as validate_main
from pyshacl.cli_rules import main as rules_main


def str_is_true(s_var: str):
    if len(s_var) > 0:
        if s_var.lower() not in ("0", "f", "n", "false", "no"):
            return True
    return False


do_server = os.getenv("PYSHACL_HTTP", "")
do_server = os.getenv("PYSHACL_SERVER", do_server)

first_arg = None if len(sys.argv) < 2 else sys.argv[1]

if first_arg is not None and str(first_arg).lower() in ('rules', '--rules'):
    rules_main(prog="python3 -m pyshacl")
elif (first_arg is not None and str(first_arg).lower() in ('serve', 'server', '--server')) or (
    do_server and str_is_true(do_server)
):
    from pyshacl.sh_http import main as http_main

    http_main()
else:
    validate_main(prog="python3 -m pyshacl")



# -*- coding: utf-8 -*-
#

# This software was developed at the National Institute of Standards
# and Technology by employees of the Federal Government in the course
# of their official duties. Pursuant to title 17 Section 105 of the
# United States Code this software is not subject to copyright
# protection and is in the public domain. NIST assumes no
# responsibility whatsoever for its use by other parties, and makes
# no guarantees, expressed or implied, about its quality,
# reliability, or any other characteristic.
#
# We would appreciate acknowledgement if the software is used.

"""
Test that values declared as the various XSD integer types are confirmed to be in their value spaces.

_PASS and _XFAIL name portions on tests in this script denote whether the input data graph should have a True or False conformance result.
"""

from typing import Set, Tuple

from rdflib import Graph, Literal, Namespace, RDF, SH, URIRef, XSD
from pyshacl import validate

EX_ONT = Namespace("http://example.com/exOnt#")

ontology_file_text = """
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix exOnt: <http://example.com/exOnt#> .

<http://example.com/exOnt> a owl:Ontology ;
    rdfs:label "An example extra-ontology file."@en .

exOnt:NumberHolder a owl:Class .

exOnt:propInteger a owl:DatatypeProperty ;
    rdfs:domain exOnt:NumberHolder ;
    rdfs:range xsd:integer .
	
exOnt:propNegativeInteger a owl:DatatypeProperty ;
    rdfs:domain exOnt:NumberHolder ;
    rdfs:range xsd:negativeInteger .
	
exOnt:propNonNegativeInteger a owl:DatatypeProperty ;
    rdfs:domain exOnt:NumberHolder ;
    rdfs:range xsd:nonNegativeInteger .
	
exOnt:propNonPositiveInteger a owl:DatatypeProperty ;
    rdfs:domain exOnt:NumberHolder ;
    rdfs:range xsd:nonPositiveInteger .
	
exOnt:propPositiveInteger a owl:DatatypeProperty ;
    rdfs:domain exOnt:NumberHolder ;
    rdfs:range xsd:positiveInteger .
"""

shacl_file_text = """
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix exShape: <http://example.com/exShape#> .
@prefix exOnt: <http://example.com/exOnt#> .

<http://example.com/exShape> a owl:Ontology ;
    rdfs:label "Example Shapes File"@en .

exShape:NumberHolderShape
    sh:property
        [
            sh:datatype xsd:integer ;
            sh:path exOnt:propInteger ;
        ],
        [
            sh:datatype xsd:negativeInteger ;
            sh:path exOnt:propNegativeInteger ;
        ],
        [
            sh:datatype xsd:nonNegativeInteger ;
            sh:path exOnt:propNonNegativeInteger
        ],
        [
            sh:datatype xsd:nonPositiveInteger ;
            sh:path exOnt:propNonPositiveInteger ;
        ],
        [
            sh:datatype xsd:positiveInteger ;
            sh:path exOnt:propPositiveInteger ;
        ] ;
    sh:targetClass exOnt:NumberHolder ;
    .
"""

data_file_text_PASS = """
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix exOnt: <http://example.com/exOnt#> .
@prefix ex: <http://example.com/ex#> .

ex:holder0 a exOnt:NumberHolder ;
    exOnt:propInteger -1 ;
    exOnt:propInteger 0 ;
    exOnt:propInteger 1 ;
    exOnt:propNegativeInteger "-1"^^xsd:negativeInteger ;
    exOnt:propNonNegativeInteger "0"^^xsd:nonNegativeInteger ;
    exOnt:propNonNegativeInteger "1"^^xsd:nonNegativeInteger ;
    exOnt:propNonPositiveInteger "-1"^^xsd:nonPositiveInteger ;
    exOnt:propNonPositiveInteger "0"^^xsd:nonPositiveInteger ;
    exOnt:propPositiveInteger "1"^^xsd:positiveInteger ;
    .
"""


def test_validate_with_ontology() -> None:
    g = Graph().parse(data=data_file_text_PASS, format='turtle')
    e = Graph().parse(data=ontology_file_text, format='turtle')
    g_len = len(g)
    res = validate(
        g, shacl_graph=shacl_file_text, shacl_graph_format='turtle', ont_graph=e, inference='both', debug=True
    )
    conforms, graph, string = res
    g_len2 = len(g)
    assert conforms
    assert g_len2 == g_len


data_file_text_XFAIL_sh_datatypes = """
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix exOnt: <http://example.com/exOnt#> .
@prefix ex: <http://example.com/ex#> .

ex:holder1 a exOnt:NumberHolder ;
    rdfs:comment "Each property except propInteger has an error.  All properties except propInteger use the wrong datatype, as Turtle infers xsd:integer on unannotated integer lexical values when none is specified."@en ;
    rdfs:seeAlso <https://www.w3.org/TR/turtle/#abbrev> ;
    exOnt:propInteger -1 ;
    exOnt:propInteger 0 ;
    exOnt:propInteger 1 ;
    exOnt:propNegativeInteger -1 ;
    exOnt:propNonNegativeInteger 0 ;
    exOnt:propNonNegativeInteger 1 ;
    exOnt:propNonPositiveInteger -1 ;
    exOnt:propNonPositiveInteger 0 ;
    exOnt:propPositiveInteger 1 ;
    .
"""


def test_validate_with_ontology_XFAIL_sh_datatypes() -> None:
    res = validate(
        data_file_text_XFAIL_sh_datatypes,
        shacl_graph=shacl_file_text,
        data_graph_format='turtle',
        shacl_graph_format='turtle',
        ont_graph=ontology_file_text,
        ont_graph_format="turtle",
        inference='both',
        debug=True,
    )
    conforms, graph, string = res
    validation_result_tally = 0
    for triple in graph.triples((None, RDF.type, SH.ValidationResult)):
        validation_result_tally += 1
    assert not conforms
    assert validation_result_tally == 6


data_file_text_XFAIL_spaces = """
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix exOnt: <http://example.com/exOnt#> .
@prefix ex: <http://example.com/ex#> .

ex:holder2 a exOnt:NumberHolder ;
    rdfs:comment "Each property has an error with the object-position node's datatype.  propInteger's has a lexical-space error.  Each other has a value-space error when considering the datatype declaration in the Literal."@en ;
    exOnt:propInteger "zero"^^xsd:integer ;
    exOnt:propNegativeInteger "0"^^xsd:negativeInteger ;
    exOnt:propNegativeInteger "1"^^xsd:negativeInteger ;
    exOnt:propNonNegativeInteger "-1"^^xsd:nonNegativeInteger ;
    exOnt:propNonPositiveInteger "1"^^xsd:nonPositiveInteger ;
    exOnt:propPositiveInteger "-1"^^xsd:positiveInteger ;
    exOnt:propPositiveInteger "0"^^xsd:positiveInteger ;
    .
"""


def test_validate_with_ontology_XFAIL_spaces() -> None:
    res = validate(
        data_file_text_XFAIL_spaces,
        shacl_graph=shacl_file_text,
        data_graph_format='turtle',
        shacl_graph_format='turtle',
        ont_graph=ontology_file_text,
        ont_graph_format="turtle",
        inference='both',
        debug=True,
    )
    conforms, graph, string = res

    expected_path_values: Set[Tuple[URIRef, Literal]] = {
        (EX_ONT.propInteger, Literal("zero", datatype=XSD.integer)),
        (EX_ONT.propNegativeInteger, Literal("0", datatype=XSD.negativeInteger)),
        (EX_ONT.propNegativeInteger, Literal("1", datatype=XSD.negativeInteger)),
        (EX_ONT.propNonNegativeInteger, Literal("-1", datatype=XSD.nonNegativeInteger)),
        (EX_ONT.propNonPositiveInteger, Literal("1", datatype=XSD.nonPositiveInteger)),
        (EX_ONT.propPositiveInteger, Literal("-1", datatype=XSD.positiveInteger)),
        (EX_ONT.propPositiveInteger, Literal("0", datatype=XSD.positiveInteger)),
    }
    computed_path_values: Set[Tuple[URIRef, Literal]] = set()

    # .triples() is used somewhat redundantly instead of using SPARQL, to avoid the possibility of any extra interpretation steps with SPARQL conversion.
    for triple0 in graph.triples((None, RDF.type, SH.ValidationResult)):
        n_validation_result = triple0[0]
        for triple1 in graph.triples((n_validation_result, SH.resultPath, None)):
            n_path = triple1[2]
            for triple2 in graph.triples((n_validation_result, SH.value, None)):
                l_value: Literal = triple2[2]
                assert isinstance(l_value, Literal)
                computed_path_values.add((n_path, l_value))
    assert not conforms
    assert expected_path_values == computed_path_values



# -*- coding: utf-8 -*-
#
"""
https://github.com/RDFLib/pySHACL/issues/213
"""
import rdflib
import pyshacl

shapes_data = '''\
@prefix ex: <http://example.org/ontology/> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .

ex:MyShape
    a sh:NodeShape ;
    sh:targetClass ex:MyFirstClass ;
    sh:property [
        rdfs:comment "This triggers as expected."@en ;
        sh:path ex:myProperty ;
        sh:minCount 1 ;
    ] ;
    sh:property [
        rdfs:comment "This does should also trigger, even when there are no myProperty values on path."@en ;
        sh:path ex:myProperty ;
        sh:qualifiedValueShape [
            sh:class ex:MySecondClass
        ] ;
        sh:qualifiedMinCount 1 ;
    ]
    .
'''

data_g_text = '''\
@prefix ex: <http://example.org/ontology/> .
@prefix kb: <http://example.org/kb/> .

kb:MyFirstClass-instance-not-ok
    a ex:MyFirstClass ;
    .

kb:MyFirstClass-instance-ok
    a ex:MyFirstClass ;
    ex:myProperty [
        a ex:MySecondClass ;
    ] ;
    .
'''


def test_213():
    shape_g = rdflib.Graph().parse(data=shapes_data, format='turtle')
    data_g = rdflib.Graph().parse(data=data_g_text, format="turtle")
    conforms, results_graph, results_text = pyshacl.validate(
        data_g,
        shacl_graph=shape_g,
        debug=True,
        meta_shacl=False,
    )
    assert not conforms
    assert "QualifiedValueShapeConstraintComponent" in results_text


if __name__ == "__main__":
    exit(test_213())



# -*- coding: utf-8 -*-
"""
https://github.com/RDFLib/pySHACL/issues/142
"""

import rdflib

from pyshacl import validate


shacl_file = """\
@prefix owl:    <http://www.w3.org/2002/07/owl#> .
@prefix rdf:    <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs:   <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh:     <http://www.w3.org/ns/shacl#> .
@prefix ex:     <http://example.com/ns#> .

ex:PersonShape
    a sh:NodeShape ;
    sh:targetClass ex:Person ;
    sh:property [
        sh:path     ex:hasPet ;
        sh:class    ex:Animal ;
        sh:minCount 1 ;
        sh:maxCount 1 ;
    ] ;
.
"""
ont_file = """\
@prefix owl:    <http://www.w3.org/2002/07/owl#> .
@prefix rdf:    <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs:   <http://www.w3.org/2000/01/rdf-schema#> .
@prefix ex:     <http://example.com/ns#> .

ex:Person a owl:Class .
ex:Animal a owl:Class .
ex:Dog    a owl:Class ; rdfs:subClassOf ex:Animal .
"""

data_file = """\
@prefix ex: <http://example.com/ns#> .
ex:Brutus a ex:Dog .
ex:Jane a ex:Person ; ex:hasPet ex:Brutus .
"""


def test_142() -> None:
    data = rdflib.Graph()
    data.parse(data=data_file, format="turtle")
    shapes = rdflib.Graph()
    shapes.parse(data=shacl_file, format="turtle")
    ont = rdflib.Graph()
    ont.parse(data=ont_file, format="turtle")
    res = validate(data, shacl_graph=shapes, ont_graph=ont)

    conforms, graph, string = res
    print(string)


if __name__ == "__main__":
    test_142()



# -*- coding: utf-8 -*-
#
"""
https://github.com/RDFLib/pySHACL/issues/98
"""
import os

from pyshacl import validate


mixed_file_text = """\
# baseURI: http://datashapes.org/sh/tests/core/complex/personexample.test
# imports: http://datashapes.org/schema
# prefix: ex

@prefix dash: <http://datashapes.org/dash#> .
@prefix ex: <http://datashapes.org/sh/tests/core/complex/personexample.test#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

<http://datashapes.org/sh/tests/core/complex/personexample.test>
  rdf:type owl:Ontology ;
  rdfs:label "Test of personexample" ;
  owl:imports <http://datashapes.org/schema> ;
.
ex:Alice
  rdf:type ex:Person ;
  ex:ssn "987-65-432A" ;
.
ex:Bob
  rdf:type ex:Person ;
  ex:ssn "123-45-6789" ;
  ex:ssn "124-35-6789" ;
.
ex:Calvin
  rdf:type ex:Person ;
  ex:birthDate "1999-09-09"^^xsd:date ;
  ex:worksFor ex:UntypedCompany ;
.

ex:PersonShape
  rdf:type sh:NodeShape ;
  sh:closed "true"^^xsd:boolean ;
  sh:ignoredProperties (
      rdf:type
    ) ;
  sh:property [
    sh:path ex:ssn ;
    sh:datatype xsd:string ;
    sh:maxCount 1 ;
    sh:pattern "^\\\\d{3}-\\\\d{2}-\\\\d{4}$" ;
    sh:message "SSN must be 3 digits - 2 digits - 4 digits."
  ] ;
  sh:property [
    sh:path ex:worksFor ;
    sh:class ex:Company ;
    sh:nodeKind sh:IRI
  ] ;
  sh:property [
      sh:path [
          sh:inversePath ex:worksFor ;
        ] ;
      sh:name "employee" ;
    ] ;
  sh:targetClass ex:Person ;
.

"""


def test_98():
    DEB_BUILD_ARCH = os.environ.get('DEB_BUILD_ARCH', None)
    DEB_HOST_ARCH = os.environ.get('DEB_HOST_ARCH', None)
    if DEB_BUILD_ARCH is not None or DEB_HOST_ARCH is not None:
        print("Cannot run owl:imports in debhelper tests.")
        assert True
        return True
    res1 = validate(
        mixed_file_text,
        shacl_graph=mixed_file_text,
        data_graph_format='turtle',
        shacl_graph_format='turtle',
        do_owl_imports=True,
        debug=True,
    )
    conforms, _, _ = res1
    assert not conforms



# -*- coding: utf-8 -*-
#
"""
https://github.com/RDFLib/pySHACL/issues/199
"""
import rdflib
from rdflib.namespace import SH
import pyshacl


shapes_data = '''\
@prefix ex: <https://example.com#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .


ex:TestConstraintComponent
  a sh:ConstraintComponent ;
  sh:parameter [
    sh:path ex:theValue ;
  ] ;
  sh:validator [
    a sh:SPARQLAskValidator ;
    sh:message "{$this}: {$theValue}" ;
    sh:ask """
      ASK {
        FILTER ($this != $this)  # Always create a validation report for this constraint component
      }""" ;
  ] .

ex:TestClass1
  a owl:Class ;
  a sh:NodeShape ;
  ex:theValue "Literal" .
ex:TestClass2
  a owl:Class ;
  a sh:NodeShape ;
  ex:theValue ex:SomeURI .
ex:TestClass3
  a owl:Class ;
  a sh:NodeShape ;
  ex:theValue _:bnode .
'''


def test_199():
    shape_g = rdflib.Graph().parse(data=shapes_data, format='turtle')
    data_g = rdflib.Graph().parse(
        data="""
    @prefix ex: <https://example.com#> .
    @prefix sh: <http://www.w3.org/ns/shacl#> .

    ex:A a ex:TestClass1 .
    ex:B a ex:TestClass2 .
    ex:C a ex:TestClass3 .
    """,
        format='turtle',
    )

    conforms, results_graph, results_text = pyshacl.validate(
        data_g,
        shacl_graph=shape_g,
        debug=True,
    )
    assert not conforms
    assert len(list(results_graph[: SH.result])) == 3


if __name__ == "__main__":
    exit(test_199())



from pyshacl import validate

data_graph = """\
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
<urn:pyshacl:datagraph> {
   <http://example.org/s1> a <http://example.org/class/1> .
   <http://example.org/s1> <http://example.org/p0> "zero"^^<http://example.org/datatypes/one> .
   <http://example.org/s1> <http://example.org/p1> "one"^^xsd:string .
   <http://example.org/s1> <http://example.org/p2> "two"@en .
   <http://example.org/s1> <http://example.org/p2> 3 .
   <http://example.org/s1> <http://example.org/p3> "4"^^<http://example.org/another/type1> .
   <http://example.org/s2> a <http://example.org/class/2> .
   <http://example.org/s2> <http://example.org/p4> "five"^^<http://example.org/another/type2> .
}"""


shapes_graph = '''\
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
<urn:pyshacl:shapes> {
    <urn:pyshacl:shapesOnt> a owl:Ontology ;
        sh:declare [
            sh:prefix "ex" ;
            sh:namespace "http://example.org/"^^xsd:anyURI ;
        ] ;
        sh:declare [
            sh:prefix "geo" ;
            sh:namespace "http://www.opengis.net/ont/geosparql#"^^xsd:anyURI ;
        ] ;
        sh:declare [
            sh:prefix "my-dt" ;
            sh:namespace "http://example.org/datatypes/"^^xsd:anyURI ;
        ] .

    <http://example.org/shapes/1> a sh:NodeShape ;
        sh:targetClass <http://example.org/class/1> , <http://example.org/class/2> ;
        sh:sparql [
            sh:prefixes <urn:pyshacl:shapesOnt> ;
            sh:select """
            SELECT DISTINCT $this ?value WHERE {
                {
                    SELECT $this ?value WHERE {
                        $this ?_p ?o .
                        FILTER (isLITERAL(?o))
                        BIND(DATATYPE(?o) as ?value)
                    }
                }
                FILTER(
                    !(
                        STRSTARTS(STR(?value), STR(xsd:)) ||
                        STRSTARTS(STR(?value), STR(rdf:)) ||
                        STRSTARTS(STR(?value), STR(my-dt:)) ||
                        STRSTARTS(STR(?value), STR(geo:))
                    )
                )
            }
            """ ;
        ] ;
        sh:message "Datatype must be from the Datatypes Namespace" .
}
'''

conforms, result_graph, result_text = \
    validate(data_graph, data_graph_format="trig", shacl_graph=shapes_graph, shacl_graph_format="trig")
print(result_text)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import os
import sys
from io import BufferedReader
from typing import Union, cast

from prettytable import PrettyTable

from pyshacl import __version__, validate
from pyshacl.errors import (
    ConstraintLoadError,
    ReportableRuntimeError,
    RuleLoadError,
    ShapeLoadError,
    ValidationFailure,
)


class ShowVersion(argparse.Action):
    def __init__(self, option_strings, dest=argparse.SUPPRESS, default=argparse.SUPPRESS, help=None):
        super(ShowVersion, self).__init__(
            option_strings=option_strings, dest=dest, default=default, nargs=0, help=help
        )

    def __call__(self, parser, namespace, values, option_string=None):
        # parser.exit() writes message to stderr before calling sys.exit()
        parser.exit(status=0, message="PySHACL Version: " + str(__version__) + "\n")


def str_is_true(s_var: str):
    if len(s_var) > 0:
        if s_var.lower() not in ("0", "f", "n", "false", "no"):
            return True
    return False


parser = argparse.ArgumentParser(description='PySHACL {} Validator command line tool.'.format(str(__version__)))
parser.add_argument(
    'data',
    metavar='DataGraph',
    help='The file or endpoint containing the Target Data Graph.',
    default=None,
    nargs='?',
)
parser.add_argument(
    '-s',
    '--shapes',
    '--shacl',
    dest='shacl',
    action='store',
    nargs='?',
    help='A file containing the SHACL Shapes Graph.',
)
parser.add_argument(
    '-e',
    '--ont-graph',
    dest='ont',
    action='store',
    nargs='?',
    help='A file path or URL to a document containing extra ontological information. '
    'RDFS and OWL definitions from this are used to inoculate the DataGraph.',
)
parser.add_argument(
    '-i',
    '--inference',
    dest='inference',
    action='store',
    default='none',
    choices=('none', 'rdfs', 'owlrl', 'both'),
    help='Choose a type of inferencing to run against the Data Graph before validating.',
)
parser.add_argument(
    '-m',
    '--metashacl',
    dest='metashacl',
    action='store_true',
    default=False,
    help='Validate the SHACL Shapes graph against the shacl-shacl Shapes Graph before validating the Data Graph.',
)
parser.add_argument(
    '-q',
    '--sparql-mode',
    dest='sparql_mode',
    action='store_true',
    default=False,
    help='Treat the DataGraph as a SPARQL endpoint, validate the graph at the SPARQL endpoint.',
)
parser.add_argument(
    '-im',
    '--imports',
    dest='imports',
    action='store_true',
    default=False,
    help='Allow import of sub-graphs defined in statements with owl:imports.',
)
parser.add_argument(
    '-a',
    '--advanced',
    dest='advanced',
    action='store_true',
    default=False,
    help='Enable features from the SHACL Advanced Features specification.',
)
parser.add_argument(
    '-j',
    '--js',
    dest='js',
    action='store_true',
    default=False,
    help='Enable features from the SHACL-JS Specification.',
)
parser.add_argument(
    '-it',
    '--iterate-rules',
    dest='iterate_rules',
    action='store_true',
    default=False,
    help="Run Shape's SHACL Rules iteratively until the data_graph reaches a steady state.",
)
parser.add_argument('--abort', dest='abort', action='store_true', default=False, help='Abort on first invalid data.')
parser.add_argument(
    '--allow-info',
    '--allow-infos',
    dest='allow_infos',
    action='store_true',
    default=False,
    help='Shapes marked with severity of Info will not cause result to be invalid.',
)
parser.add_argument(
    '-w',
    '--allow-warning',
    '--allow-warnings',
    dest='allow_warnings',
    action='store_true',
    default=False,
    help='Shapes marked with severity of Warning or Info will not cause result to be invalid.',
)
parser.add_argument(
    '--max-depth',
    dest='max_depth',
    action='store',
    nargs='?',
    type=int,
    help="The maximum number of SHACL shapes \"deep\" that the validator can go before reaching an \"endpoint\" constraint.",
)
parser.add_argument(
    '-d',
    '--debug',
    dest='debug',
    action='store_true',
    default=False,
    help='Output additional verbose runtime messages.',
)
parser.add_argument(
    '--focus',
    dest='focus',
    action='store',
    help='Optional IRIs of focus nodes from the DataGraph, the shapes will validate only these node. Comma-separated list.',
    nargs="?",
    default=None,
)
parser.add_argument(
    '--shape',
    dest='shape',
    action='store',
    help='Optional IRIs of a NodeShape or PropertyShape from the SHACL ShapesGraph, only these shapes will be used to validate the DataGraph. Comma-separated list.',
    nargs="?",
    default=None,
)
parser.add_argument(
    '-f',
    '--format',
    dest='format',
    action='store',
    help='Choose an output format. Default is "human".',
    default='human',
    choices=('human', 'table', 'turtle', 'xml', 'json-ld', 'nt', 'n3'),
)
parser.add_argument(
    '-df',
    '--data-file-format',
    dest='data_file_format',
    action='store',
    help='Explicitly state the RDF File format of the input DataGraph file. Default="auto".',
    default='auto',
    choices=('auto', 'turtle', 'xml', 'json-ld', 'nt', 'n3'),
)
parser.add_argument(
    '-sf',
    '--shacl-file-format',
    dest='shacl_file_format',
    action='store',
    help='Explicitly state the RDF File format of the input SHACL file. Default="auto".',
    default='auto',
    choices=('auto', 'turtle', 'xml', 'json-ld', 'nt', 'n3'),
)
parser.add_argument(
    '-ef',
    '--ont-file-format',
    dest='ont_file_format',
    action='store',
    help='Explicitly state the RDF File format of the extra ontology file. Default="auto".',
    default='auto',
    choices=('auto', 'turtle', 'xml', 'json-ld', 'nt', 'n3'),
)
parser.add_argument('-V', '--version', action=ShowVersion, help='Show PySHACL version and exit.')
parser.add_argument(
    '-o',
    '--output',
    dest='output',
    nargs='?',
    type=argparse.FileType('w'),
    help='Send output to a file (defaults to stdout).',
    default=sys.stdout,
)
parser.add_argument(
    '--rules',
    help='Ignore validation options, run PySHACL in Rules Expansion mode. Same as `pyshacl_rules`.',
    action='store_true',
    dest='do_rules',
    default=False,
)
parser.add_argument(
    '--server',
    help='Ignore all the rest of the options, start the HTTP Server. Same as `pyshacl_server`.',
    action='store_true',
    dest='server',
    default=False,
)
# parser.add_argument('-h', '--help', action="help", help='Show this help text.')


def main(prog: Union[str, None] = None) -> None:
    if prog is not None and len(prog) > 0:
        parser.prog = prog
    do_server = os.getenv("PYSHACL_HTTP", "")
    do_server = os.getenv("PYSHACL_SERVER", do_server)
    if do_server:
        args = argparse.Namespace()
    else:
        args = parser.parse_args()
    if str_is_true(do_server) or args.server:
        from pyshacl.sh_http import main as http_main

        # http_main calls sys.exit(0) and never returns
        http_main()
    if args.do_rules:
        from pyshacl.cli_rules import main as rules_main

        # rules_main calls sys.exit(0) and never returns
        rules_main()
    if not args.data:
        # No datafile give, and not starting in server mode.
        sys.stderr.write('Input Error. No DataGraph file or endpoint supplied.\n')
        parser.print_usage(sys.stderr)
        sys.exit(1)
    validator_kwargs = {'debug': args.debug}
    data_file = None
    data_graph: Union[BufferedReader, str]
    if args.sparql_mode is not None and args.sparql_mode is True:
        endpoint = str(args.data).strip()
        if not endpoint.lower().startswith("http:") and not endpoint.lower().startswith("https:"):
            sys.stderr.write("Input Error. SPARQL Endpoint must start with http:// or https://.\n")
            sys.exit(1)
        data_graph = endpoint
        validator_kwargs['sparql_mode'] = True
    else:
        try:
            data_file = open(args.data, 'rb')
        except FileNotFoundError:
            sys.stderr.write('Input Error. DataGraph file not found.\n')
            sys.exit(1)
        except PermissionError:
            sys.stderr.write('Input Error. DataGraph file not readable.\n')
            sys.exit(1)
        else:
            # NOTE: This cast is not necessary in Python >= 3.10.
            data_graph = cast(BufferedReader, data_file)
    if args.shacl is not None:
        validator_kwargs['shacl_graph'] = args.shacl
    if args.ont is not None:
        validator_kwargs['ont_graph'] = args.ont
    if args.format not in ['human', 'table']:
        validator_kwargs['serialize_report_graph'] = args.format
    if args.inference != 'none':
        validator_kwargs['inference'] = args.inference
    if args.imports:
        validator_kwargs['do_owl_imports'] = True
    if args.metashacl:
        validator_kwargs['meta_shacl'] = True
    if args.advanced:
        validator_kwargs['advanced'] = True
    if args.js:
        validator_kwargs['js'] = True
    if args.focus:
        validator_kwargs['focus_nodes'] = [_f.strip() for _f in args.focus.split(',')]
    if args.shape:
        validator_kwargs['use_shapes'] = [_s.strip() for _s in args.shape.split(',')]
    if args.iterate_rules:
        if not args.advanced:
            sys.stderr.write("Iterate-Rules option only works when you enable Advanced Mode.\n")
        else:
            validator_kwargs['iterate_rules'] = True
    if args.abort:
        validator_kwargs['abort_on_first'] = True
    if args.allow_infos:
        validator_kwargs['allow_infos'] = True
    if args.allow_warnings:
        validator_kwargs['allow_warnings'] = True
    if args.shacl_file_format:
        _f: str = args.shacl_file_format
        if _f != "auto":
            validator_kwargs['shacl_graph_format'] = _f
    if args.ont_file_format:
        _f = args.ont_file_format
        if _f != "auto":
            validator_kwargs['ont_graph_format'] = _f
    if args.data_file_format:
        _f = args.data_file_format
        if _f != "auto":
            validator_kwargs['data_graph_format'] = _f
    exit_code: Union[int, None] = None
    try:
        is_conform, v_graph, v_text = validate(data_graph, **validator_kwargs)
        if isinstance(v_graph, BaseException):
            raise v_graph
    except ValidationFailure as vf:
        args.output.write("Validator generated a Validation Failure result:\n")
        args.output.write(str(vf.message))
        args.output.write("\n")
        exit_code = 1
    except ShapeLoadError as sle:
        sys.stderr.write("Validator encountered a Shape Load Error:\n")
        sys.stderr.write(str(sle))
        exit_code = 2
    except ConstraintLoadError as cle:
        sys.stderr.write("Validator encountered a Constraint Load Error:\n")
        sys.stderr.write(str(cle))
        exit_code = 2
    except RuleLoadError as rle:
        sys.stderr.write("Validator encountered a Rule Load Error:\n")
        sys.stderr.write(str(rle))
        exit_code = 2
    except ReportableRuntimeError as rre:
        sys.stderr.write("Validator encountered a Runtime Error:\n")
        sys.stderr.write(str(rre.message))
        sys.stderr.write("\nIf you believe this is a bug in pyshacl, open an Issue on the pyshacl github page.\n")
        exit_code = 2
    except NotImplementedError as nie:
        sys.stderr.write("Validator feature is not implemented:\n")
        if len(nie.args) > 0:
            sys.stderr.write(str(nie.args[0]))
        else:
            sys.stderr.write("No message provided.")
        sys.stderr.write("\nIf your use-case requires this feature, open an Issue on the pyshacl github page.\n")
        exit_code = 3
    except RuntimeError as re:
        import traceback

        traceback.print_tb(re.__traceback__)
        sys.stderr.write(
            "\n\nValidator encountered a Runtime Error. Please report this to the PySHACL issue tracker.\n"
        )
        exit_code = 2
    finally:
        if data_file is not None:
            try:
                data_file.close()
            except Exception as e:
                sys.stderr.write("Error closing data file:\n")
                sys.stderr.write(str(e))
        if exit_code is not None:
            sys.exit(exit_code)
    if args.format == 'human':
        args.output.write(v_text)
    elif args.format == 'table':
        t1 = PrettyTable()
        t1.field_names = ["Conforms"]
        t1.align = "c"
        t1.add_row([is_conform])
        args.output.write(str(t1))
        args.output.write('\n\n')

        def col_widther(s, w):
            """Split strings to a given width for table"""
            s2 = []
            i = 0
            while i < len(s):
                s2.append(s[i : i + w])
                i += w
            return '\n'.join(s2)

        if not is_conform:
            from rdflib import Graph
            from rdflib.namespace import SH

            t2 = PrettyTable()
            t2.field_names = ['No.', 'Severity', 'Focus Node', 'Result Path', 'Message', 'Component', 'Shape', 'Value']
            t2.align = "l"

            assert isinstance(v_graph, Graph)
            for i, o in enumerate(v_graph.objects(None, SH.result)):
                r = {}
                for o2 in v_graph.predicate_objects(o):
                    r[o2[0]] = str(col_widther(str(o2[1]).replace(f'{SH}', ''), 25))  # max col width 30 chars
                t2.add_row(
                    [
                        i + 1,
                        r[SH.resultSeverity],
                        r[SH.focusNode],
                        r[SH.resultPath] if r.get(SH.resultPath) is not None else '-',
                        r[SH.resultMessage] if r.get(SH.resultMessage) is not None else '-',
                        r[SH.sourceConstraintComponent],
                        r[SH.sourceShape],
                        r[SH.value] if r.get(SH.value) is not None else '-',
                    ]
                )
                t2.add_row(['', '', '', '', '', '', '', ''])
            args.output.write(str(t2))
    else:
        if isinstance(v_graph, bytes):
            v_graph = v_graph.decode('utf-8')
        args.output.write(v_graph)
    args.output.close()
    sys.exit(0 if is_conform else 1)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
#
"""
https://github.com/RDFLib/pySHACL/issues/9
"""
from pyshacl import validate

shacl_file_text = """
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix ex: <http://example.com/ex#> .

ex:Parent a rdfs:Class ;
    rdfs:isDefinedBy ex: ;
    rdfs:comment "The parent class"@en ;
    rdfs:subClassOf owl:Thing .

ex:ParentShape a sh:NodeShape ;
    rdfs:isDefinedBy ex: ;
    sh:property [
        sh:datatype xsd:string ;
        sh:path ex:name ;
        sh:maxCount 1 ;
        sh:minCount 1 ;
    ] ;
    sh:closed true ;
    sh:ignoredProperties ( rdf:type ) ;
    sh:targetClass ex:Parent .
"""

data_file_text = """
{
    "@context": {
        "@vocab": "http://example.com/ex#"
    },
    "@type": "Parent",
    "name": "Father",
    "dummy": "Dummy value"
}
"""


def test_009():
    res = validate(
        data_file_text,
        shacl_graph=shacl_file_text,
        data_graph_format='json-ld',
        shacl_graph_format='turtle',
        inference='both',
        debug=True,
    )
    conforms, graph, string = res
    assert not conforms



import logging
from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Optional

import rdflib

from pyshacl.errors import ReportableRuntimeError

if TYPE_CHECKING:
    from rdflib.term import URIRef

    from pyshacl.pytypes import GraphLike


class PySHACLRunType(metaclass=ABCMeta):
    __slots__ = ()

    @abstractmethod
    def run(self):
        raise NotImplementedError()  # pragma: no cover

    @classmethod
    def _run_pre_inference(
        cls,
        target_graph: 'GraphLike',
        inference_option: str,
        destination_graph_identifier: Optional['URIRef'] = None,
        logger: Optional[logging.Logger] = None,
    ):
        """
        Note, this is the OWL/RDFS pre-inference,
        it is not the Advanced Spec SHACL-Rule inferencing step.
        :param target_graph:
        :type target_graph: rdflib.Graph|rdflib.ConjunctiveGraph|rdflib.Dataset
        :param inference_option:
        :type inference_option: str
        :return:
        :rtype: NoneType
        """
        # Lazy import owlrl
        import owlrl

        from .inference import CustomRDFSOWLRLSemantics, CustomRDFSSemantics

        if logger is None:
            logger = logging.getLogger(__name__)
        try:
            if inference_option == 'rdfs':
                inferencer = owlrl.DeductiveClosure(CustomRDFSSemantics)
            elif inference_option == 'owlrl':
                inferencer = owlrl.DeductiveClosure(owlrl.OWLRL_Semantics)
            elif inference_option == 'both' or inference_option == 'all' or inference_option == 'rdfsowlrl':
                inferencer = owlrl.DeductiveClosure(CustomRDFSOWLRLSemantics)
            else:
                raise ReportableRuntimeError("Don't know how to do '{}' type inferencing.".format(inference_option))
        except Exception as e:  # pragma: no cover
            logger.error("Error during creation of OWL-RL Deductive Closure")
            if isinstance(e, ReportableRuntimeError):
                raise e
            raise ReportableRuntimeError(
                "Error during creation of OWL-RL Deductive Closure\n{}".format(str(e.args[0]))
            )
        if isinstance(target_graph, (rdflib.Dataset, rdflib.ConjunctiveGraph)):
            target_graph.default_union = True
            if destination_graph_identifier is not None:
                destination_graph = target_graph.get_context(destination_graph_identifier)
            else:
                destination_graph = target_graph.default_context
        else:
            destination_graph = None
        try:
            inferencer.expand(target_graph, destination=destination_graph)
        except Exception as e:  # pragma: no cover
            raise
            logger.error("Error while running OWL-RL Deductive Closure")
            raise ReportableRuntimeError("Error while running OWL-RL Deductive Closure\n{}".format(str(e.args[0])))
